from ..storages import LocalMemoryStorage
from ..storages import Storage
from unittest import mock
import pytest


@pytest.fixture
def local_storage():
    return LocalMemoryStorage()


@mock.patch('cq.storages.genuuid', lambda: 'EVENT_ID')
@mock.patch('cq.storages.publish')
def test_store(publish):
    storage = Storage()
    with mock.patch.object(storage, 'create_event', return_value='EVENT') as create_event,\
         mock.patch.object(storage, 'append') as append:
        storage.store('User.Registered', 'aabbcc', {'name': 'joe'})
        create_event.assert_called_once_with(
            id='EVENT_ID',
            name='User.Registered',
            aggregate_id='aabbcc',
            data={'name': 'joe'},
            ts=None,
        )
        append.assert_called_once_with('EVENT')
        publish.assert_called_once_with('EVENT')


def test_local__append(local_storage):
    local_storage.store('User.Registered', 'JOE_ID', {'name': 'joe'}, 'TS')
    local_storage.store('User.Registered', 'JANE_ID', {'name': 'jane'}, 'TS')
    local_storage.store('User.Activated', 'JOE_ID', ts='TS')

    assert [(e.name, e.aggregate_id, e.data, e.ts) for e in local_storage.events] == [
        ('User.Registered', 'JOE_ID', {'name': 'joe'}, 'TS'),
        ('User.Registered', 'JANE_ID', {'name': 'jane'}, 'TS'),
        ('User.Activated', 'JOE_ID', None, 'TS'),
    ]


def test_local__get_events(local_storage):
    local_storage.store('User.Registered', 'JOE_ID', {'name': 'joe'}, 'TS')
    local_storage.store('User.Registered', 'JANE_ID', {'name': 'jane'}, 'TS')
    local_storage.store('User.Activated', 'JOE_ID', {'name': 'joe'}, 'TS')

    assert [(e.name, e.aggregate_id) for e in local_storage.get_events('JOE_ID')] == [
        ('User.Registered', 'JOE_ID'),
        ('User.Activated', 'JOE_ID'),
    ]
    assert [(e.name, e.aggregate_id) for e in local_storage.get_events('JANE_ID')] == [
        ('User.Registered', 'JANE_ID'),
    ]


def test_local__book_unique(local_storage):
    local_storage.book_unique('user_email', 'joe@doe.com', 'JOE_ID')
    assert local_storage.has_unique('user_email', 'joe@doe.com') is True
    assert local_storage.get_unique('user_email', 'joe@doe.com') == 'JOE_ID'


def test_local__book_unique_fails_for_duplicate(local_storage):
    local_storage.book_unique('user_email', 'joe@doe.com', 'JOE_ID')

    with pytest.raises(Storage.DuplicatedItemError):
        local_storage.book_unique('user_email', 'joe@doe.com', 'JOE_ID2')

    # first value should not be overridden
    assert local_storage.get_unique('user_email', 'joe@doe.com') == 'JOE_ID'



from .exceptions import SesError
from .events import Event
from .genuuid import genuuid
from .handlers import publish
from collections import defaultdict
from collections import namedtuple


UniqueItem = namedtuple('UniqueItem', '')


class Storage:
    class DuplicatedItemError(SesError):
        pass

    class DoesNotExist(SesError):
        pass

    def store(self, name, aggregate_id, data=None, ts=None):
        event = self.create_event(
            id=genuuid(),
            name=name,
            aggregate_id=aggregate_id,
            data=data,
            ts=ts,
        )
        self.append(event)
        publish(event)
        return event

    def create_event(self, id, name, aggregate_id, data=None, ts=None):
        return Event(
            id=id,
            name=name,
            aggregate_id=aggregate_id,
            data=data,
            ts=ts,
        )

    def append(self, event):
        raise NotImplementedError

    def get_events(self, aggregate_id):
        raise NotImplementedError

    def book_unique(self, namespace, value, aggregate_id=None):
        raise NotImplementedError

    def get_unique(self, namespace, value):
        raise NotImplementedError

    def has_unique(self, namespace, value):
        raise NotImplementedError


class LocalMemoryStorage(Storage):

    def __init__(self):
        self.events = []
        self.uniques = defaultdict(dict)

    def append(self, event):
        self.events.append(event)
        return event

    def get_events(self, aggregate_id):
        return [e for e in self.events if e.aggregate_id == aggregate_id]

    def book_unique(self, namespace, value, aggregate_id=None):
        if value in self.uniques[namespace]:
            raise Storage.DuplicatedItemError('%s:%s already exists' % (namespace, value))
        else:
            self.uniques[namespace][value] = aggregate_id

    def get_unique(self, namespace, value):
        if value in self.uniques[namespace]:
            return self.uniques[namespace][value]
        else:
            raise self.DoesNotExist('%s:%s was not set' % (namespace, value))

    def has_unique(self, namespace, value):
        return value in self.uniques[namespace]



from collections import defaultdict


handlers_registry = defaultdict(set)


def register_handler(event_name):
    def wrap(handler):
        handlers_registry[event_name].add(handler)

        def inner(event, aggregate):
            return
        return handler

    return wrap


def publish(event):
    for handler in get_handlers(event):
        handler(event)


def get_handlers(event):
    return handlers_registry[event.name]



import datetime


class Event:
    def __init__(self, id, name, aggregate_id, data=None, ts=None):
        self.id = id
        self.name = name
        self.aggregate_id = aggregate_id
        self.data = data
        self.ts = ts or datetime.datetime.now()

    def __str__(self):
        return '%s | %s | %s' % (self.name, self.aggregate_id, self.ts)



from django.db import models
from django.utils import timezone
from cq.genuuid import genuuid
import jsonfield


class Event(models.Model):
    id = models.CharField(max_length=128, primary_key=True, default=genuuid)
    ts = models.DateTimeField(default=timezone.now)
    name = models.CharField(max_length=128, db_index=True)
    aggregate_id = models.CharField(max_length=128, db_index=True)
    data = jsonfield.JSONField(null=True)

    def __str__(self):
        return '%s | %s | %s' % (self.name, self.aggregate_id, self.ts)


class UniqueItem(models.Model):
    namespace = models.CharField(max_length=128)
    value = models.CharField(max_length=255)
    aggregate_id = models.CharField(max_length=128)

    class Meta:
        unique_together = ('namespace', 'value')

    def __str__(self):
        return '%s | %s' % (self.namespace, self.value)



from cq.contrib.django.models import Event as EventModel
from cq.contrib.django.models import UniqueItem
from cq.events import Event
from cq.storages import Storage
from django.db import IntegrityError


class DjangoStorage(Storage):

    def append(self, event):
        obj = to_model(event)
        obj.save()
        return obj

    def get_events(self, aggregate_id):
        qs = EventModel.objects.filter(aggregate_id=aggregate_id).order_by('ts')
        return (from_model(e) for e in qs)

    def book_unique(self, namespace, value, aggregate_id):
        try:
            UniqueItem.objects.create(namespace=namespace, value=value, aggregate_id=aggregate_id)
        except IntegrityError:
            raise Storage.DuplicatedItemError('%s:%s already exists' % (namespace, value))

    def get_unique(self, namespace, value):
        try:
            return UniqueItem.objects.get(namespace=namespace, value=value).aggregate_id
        except UniqueItem.DoesNotExist:
            raise self.DoesNotExist('%s:%s was not set' % (namespace, value))

    def has_unique(self, namespace, value):
        return UniqueItem.objects.filter(namespace=namespace, value=value).exists()


def to_model(event):
    return EventModel(
        id=event.id,
        name=event.name,
        aggregate_id=event.aggregate_id,
        data=event.data,
    )


def from_model(instance):
    return Event(
        id=instance.id,
        name=instance.name,
        aggregate_id=instance.aggregate_id,
        data=instance.data,
        ts=instance.ts,
    )



class Aggregate:
    mutators = {}

    def __init__(self, id):
        self.id = id
        self.version = 1

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.id)

    def mutate(self, event):
        mutator = self.get_mutator(event.name)
        mutator(self, event, event.data)
        self.version += 1

    def get_mutator(self, name):
        try:
            return self.__class__.mutators[name]
        except KeyError:
            msg = '%s has no mutator function registered for event type %r. ' % (self, name)
            msg += 'Please register method using @aggregates.register_mutator decorator. '
            registered_actions = ', '.join(repr(a) for a in self.__class__.mutators)
            msg += 'Currently registered actions: %s' % (registered_actions or '-')
            raise NotImplementedError(msg)

    @classmethod
    def get_name(cls):
        return cls.__name__


def register_mutator(aggregate_class, event_name):

    def outer(method):
        if event_name in aggregate_class.mutators:
            msg = "Mutator for action %s is already registered for %s" % (event_name, aggregate_class)
            raise RuntimeError(msg)
        else:
            aggregate_class.mutators[event_name] = method
        return method

    return outer


class Repository:
    aggregate_name = None
    aggregate_class = None

    class DoesNotExist(Exception):
        pass

    class DuplicateAggregateError(Exception):
        pass

    def __init__(self, storage):
        self.storage = storage
        if self.aggregate_class is None:
            msg = "Repository must define aggregate. Please set %s.aggregate to Aggregate subclass"
            raise RuntimeError(msg % self)

    def store(self, name, aggregate_id, data=None):
        return self.storage.store(name, aggregate_id, data)

    def get_events(self, aggregate_id):
        return self.storage.get_events(aggregate_id)

    def get_aggregate(self, aggregate_id):
        aggregate = self.aggregate_class(aggregate_id)
        for event in self.get_events(aggregate_id):
            aggregate.mutate(event)
        return aggregate

    def get_aggregate_name(self):
        if self.aggregate_name:
            return self.aggregate_name
        return self.aggregate_class.get_name()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-11-05 06:52
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import jsonfield.fields
import cq.genuuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.CharField(default=cq.genuuid.genuuid, max_length=128, primary_key=True, serialize=False)),
                ('ts', models.DateTimeField(default=django.utils.timezone.now)),
                ('name', models.CharField(db_index=True, max_length=128)),
                ('aggregate_id', models.CharField(db_index=True, max_length=128)),
                ('data', jsonfield.fields.JSONField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name='UniqueItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('namespace', models.CharField(max_length=128)),
                ('value', models.CharField(max_length=255)),
                ('aggregate_id', models.CharField(max_length=128)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='uniqueitem',
            unique_together=set([('namespace', 'value')]),
        ),
    ]



from django.utils.module_loading import import_string
import os


DEFAULT_STORAGE_CLASS = import_string('cq.storages.LocalMemoryStorage')
DB_URI = os.environ.get('DB_URI')
ENGINE_ECHO = 'SES_SQLALCHEMY_ENGINE_ECHO' in os.environ


# if Django is configured default storage would be Django specific one
# unless user explicitly set SES_DEFAULT_STORAGE_CLASS at settings
# pointing to another storage
try:
    from django.conf import settings
    if settings.configured:
        cls = getattr(settings, 'SES_DEFAULT_STORAGE_CLASS',
                      'cq.contrib.django.storages.DjangoStorage')
        DEFAULT_STORAGE_CLASS = import_string(cls)
except ImportError:
    # we are not in a Django context, do nothing
    pass



from . import exceptions
from . import genuuid
from . import settings
from .aggregates import Repository


class BaseApp:
    storage_class = None
    storage_kwargs = {}

    def __init__(self):
        self.storage = self.get_storage()

    def get_storage(self):
        storage_class = self.get_storage_class()
        storage_kwargs = self.get_storage_kwargs()
        return storage_class(**storage_kwargs)

    def get_storage_class(self):
        if self.storage_class is None:
            if settings.DEFAULT_STORAGE_CLASS:
                return settings.DEFAULT_STORAGE_CLASS
            msg = "either storage_class must be set or override get_storage method"
            raise exceptions.ImproperlyConfigured(msg)
        return self.storage_class

    def get_storage_kwargs(self):
        return self.storage_kwargs

    def genuuid(self):
        return genuuid.genuuid()

    def get_repo_for_aggregate(self, aggregate_class):
        name = '%sRepository' % aggregate_class
        repo_class = type(name, (Repository,), {'aggregate_class': aggregate_class})
        return repo_class(storage=self.storage)



from setuptools import setup
import os
import os.path


def get_packages(top):
    packages = []
    for dirname, subdirs, filenames in os.walk(top):
        initpy = os.path.join(dirname, '__init__.py')
        is_python_package = os.path.isfile(initpy)
        if is_python_package:
            packages.append(dirname)
    return packages


setup(
    name='cq',
    version='0.8',
    url='https://github.com/lukaszb/cq',
    license='MIT',
    description='simple cqrs implementation',
    author='Lukasz Balcerzak',
    author_email='lukaszbalcerzak@gmail.com',
    zip_safe=False,
    packages=get_packages('cq'),
    include_package_data=True,
    install_requires=[
        'jsonfield',  # TODO: only needed django, make it optional
    ],
)



from cq import settings
from cq.events import Event
from cq.genuuid import genuuid
from cq.storages import Storage
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative.api import declarative_base
import json
import sqlalchemy


Base = declarative_base()


class EventModel(Base):
    __tablename__ = 'cq_event'

    id = sqlalchemy.Column(sqlalchemy.String(128), primary_key=True, default=genuuid)
    name = sqlalchemy.Column(sqlalchemy.String(255), index=True)
    aggregate_id = sqlalchemy.Column(sqlalchemy.String(255), index=True)
    data = sqlalchemy.Column(sqlalchemy.Text(), default='{}')
    ts = sqlalchemy.Column(sqlalchemy.DateTime(), index=True)


class SqlAlchemyStorage(Storage):

    def __init__(self, db_uri=None, echo=None):
        self.db_uri = db_uri
        self.echo = settings.ENGINE_ECHO

    def append(self, event):
        obj = to_model(event)
        session = self.get_session()
        session.add(obj)
        session.commit()
        return obj

    def get_events(self, aggregate_id):
        session = self.get_session()
        # TODO: should be ordered by version, not ts (otoh ts should also work)
        query = session.query(EventModel).filter(EventModel.aggregate_id==aggregate_id).order_by(EventModel.ts)
        return (from_model(e) for e in query)

    def book_unique(self, namespace, value, aggregate_id=None):
        raise NotImplementedError

    def get_unique(self, namespace, value):
        raise NotImplementedError

    def has_unique(self, namespace, value):
        raise NotImplementedError

    def get_session(self):
        Session = self.get_sessionmaker()
        return Session()

    def get_sessionmaker(self):
        if not hasattr(self, '_sessionmaker'):
            engine = self.get_engine()
            self._sessionmaker = sessionmaker(bind=engine)
        return self._sessionmaker

    def get_engine(self):
        db_uri = self.get_db_uri()
        if not hasattr(self, '_engine'):
            self._engine = sqlalchemy.create_engine(db_uri, echo=self.echo)
            Base.metadata.create_all(self._engine)
        return self._engine

    def get_db_uri(self):
        if self.db_uri:
            return self.db_uri
        if settings.DB_URI is None:
            raise RuntimeError("DB_URI environment variable must be set")
        else:
            return settings.DB_URI


def to_model(event):
    return EventModel(
        id=event.id,
        name=event.name,
        aggregate_id=event.aggregate_id,
        data=json.dumps(event.data),
    )


def from_model(instance):
    return Event(
        id=instance.id,
        name=instance.name,
        aggregate_id=instance.aggregate_id,
        data=json.loads(instance.data),
        ts=instance.ts,
    )
